"""
Time series module for Python
==================================
giottotime is a Python module integrating time series and machine learning.

See URL TO GIOTTOTIME for complete documentation.
"""

from giottotime._version import __version__


__all__ = [
    "causality_tests",
    "feature_creation",
    "loss_functions",
    "model_selection",
    "models",
    "time_series_preparation",
]
